#include <queue>
#include <string>
#include <strstream>
#include <pthread.h>
#include "v4lxif.h"

#ifndef _VCR_CAPPROC_H_
#define _VCR_CAPPROC_H_

struct chunk
{
    char* data;
    double timestamp;
};    
class CaptureProcess
{
public:
    enum Sound_Freqs
    {
	NoAudio	=-1,
	F44	=0,
	F22	=1,
	F11	=2
    };
    enum Sample_Sizes
    {
	S16	=0,
	S8	=1
    };
    enum Sound_Chans
    {
	Mono	=0,
	Stereo	=1
    };
    enum Resolutions
    {
	W384	=0,
	W320	=1,
	W192	=2,
	W160	=3
    };    
protected:
    static void* vidcap(void*);
    static void* audcap(void*);
    static void* writer(void*);
protected:
    v4lxif* m_v4l;
    pthread_t m_vidc;
    pthread_t m_audc;
    pthread_t m_writer;
    int m_quit;
    std::queue<chunk> m_vidq;
    std::queue<chunk> m_audq;
    int cnt;
    int cap_drop;
    int comp_drop;
    long long starttime;

    std::strstream* messenger;
    std::strstream def_mess;

    std::string filename;
    int segment_size;		
    int compressor;		
    int quality;		
    int keyframes;
    int frequency;
    int samplesize;
    int chan;
    int res_w;
    int res_h;
    int timelimit;	
    int sizelimit;
    float fps;	

    int segment_flag;

    int vid_clear;
    int aud_clear;
    
    void Create(v4lxif* v4l, std::string filename, int segment_size,	
	int compressor, int quality, int keyframes,
        enum Sound_Freqs frequency, enum Sample_Sizes samplesize,
        enum Sound_Chans chan, enum Resolutions res, int timelimit,		
	int sizelimit, float fps );

public:
    CaptureProcess(v4lxif* v4l,		// v4lx interface pointer
	std::string filename,
	int segment_size,		// Segmentation is implicit finishing one
    				// file and starting next one when old file
    				// grows to specified size
    				//Kbytes
    				//-1 for no segmentation
	int compressor,		// fourcc
	int quality,		//0..10000
	int keyframes,		// compressor-dependent
				// for IV 5.0 Quick it's ignored
	enum Sound_Freqs frequency,		// frequency ID
				// Available:
				//-1 for no audio
				//0 for 44100 Hz
				//1 for 22050 Hz
				//2 for 11025 Hz
        enum Sample_Sizes samplesize,		//0 for 16 bit
				//1 for 8 bit
	enum Sound_Chans chan,
	enum Resolutions res,
	int timelimit,		//in seconds. -1 if no limit
	int sizelimit,		//in Kbytes. -1 if no limit
        float fps): messenger(&def_mess), segment_flag(0), vid_clear(0), aud_clear(0)
    {
    	Create(v4l, filename, 
	    segment_size, compressor, quality, keyframes,		
	    frequency, samplesize, chan, res, timelimit, sizelimit, fps);
    }
    CaptureProcess(int v4l,	// initialized v4l device
	std::string filename,
        int segment_size,		
	int compressor,		
        int quality,		
	int keyframes,		
        enum Sound_Freqs frequency,	
	enum Sample_Sizes samplesize,
        enum Sound_Chans chan,
	enum Resolutions res,
	int timelimit,		
        int sizelimit,
	float fps): messenger(&def_mess), segment_flag(0), vid_clear(0), aud_clear(0)
    {
	Create((v4lxif*)(new v4l1if(v4l)), filename, 
	    segment_size, compressor, quality, keyframes,		
	    frequency, samplesize, chan, res, timelimit, sizelimit, fps);
    }
    ~CaptureProcess();
    void getState(int& processed, int& cap_drop, int& comp_drop, int& finished)
    {
	processed=cnt;
	cap_drop=this->cap_drop;
	comp_drop=this->comp_drop;
	finished=m_quit;
    }
    void setMessenger(std::strstream& mess)
    {
	messenger=&mess;
    }
    void Segment()
    {
	segment_flag=1;
//	vid_clear=1;
//	aud_clear=1;
    }	
};
#endif /* _VCR_CAPPROC_H_ */
